@@ -15,15 +15,25 @@
 /*  `final' properties to an input string of character codes for the Arabic */
 /*  script.                                                                 */
 /*                                                                          */
+/*  This file is part of the FreeType project, and may only be used         */
+/*  modified and distributed under the terms of the FreeType project        */
+/*  license, LICENSE.TXT.  By continuing to use, modify, or distribute      */
+/*  this file you indicate that you have read the license and               */
+/*  understand and accept it fully.                                         */
+/*                                                                          */
+/* The file LICENSE.TXT can be found in the Pango distribution as           */
+/* pango/opentype/FT-license.txt                                            */
 /****************************************************************************/
 
+#include <config.h>
+
 #include "arabic-ot.h"
 
 
 /*
  *
  *     Here a table of the joining classes for characters in the range
- *     U+0620 - U+06FF.
+ *     U+0620 - U+06FF and U+0750 - U+077F.
  *
  *    The following character also has a joining class:
  *
@@ -56,8 +66,8 @@ static const joining_class  arabic[] =
   /* U+0650 */
   transparent, transparent, transparent, transparent,
   transparent, transparent, transparent, transparent,
-  none, none, none, none,
-  none, none, none, none,
+  transparent, transparent, transparent, transparent,
+  transparent, transparent, transparent, none,
 
   /* U+0660 */
   none, none, none, none,
@@ -120,6 +130,27 @@ static const joining_class  arabic[] =
   dual, none, none, dual
 };
 
+static const joining_class  arabic_supplement[] =
+{
+  /* U+0750 */
+  dual, dual, dual, dual,
+  dual, dual, dual, dual,
+  dual, right, right, right,
+  dual, dual, dual, dual,
+
+  /* U+0760 */
+  dual, dual, dual, dual,
+  dual, dual, dual, dual,
+  dual, dual, dual, right,
+  right, dual, none, none,
+
+  /* U+0770 */
+  none, none, none, none,
+  none, none, none, none,
+  none, none, none, none,
+  none, none, none, none
+};
+
 #if 0
 struct cgc_
 {
@@ -265,18 +296,18 @@ static joining_class  Get_Joining_Class (gunichar*   string,
 
       if (pos >= length)
         return none;
-      
-      if (string[pos] < 0x0620 ||
-	  string[pos] >= 0x0700)
-	{
-	  if (string[pos] == 0x200D)
-	    return causing;
-	  else
-	    return none;
-	}
+
+      if (string[pos] >= 0x0620 &&
+          string[pos] < 0x0700)
+        j = arabic[string[pos] - 0x0620];
+      else if (string[pos] >= 0x0750 &&
+          string[pos] < 0x0780)
+        j = arabic_supplement[string[pos] - 0x0750];
+      else if (string[pos] == 0x200D)
+        return causing;
       else
-        j =  arabic[string[pos] - 0x0620];
-      
+        return none;
+
       if (!direction || j != transparent)
         return j;
     }
