@@ -21,7 +21,7 @@
  * Boston, MA 02111-1307, USA.
  */
 
-#include "config.h"
+#include <config.h>
 
 #include <ctype.h>
 #include <stdio.h>
@@ -32,43 +32,14 @@
 #include "pango-fontmap.h"
 #include "pango-utils.h"
 #include "pangowin32-private.h"
+#include "modules.h"
 
-#define PANGO_TYPE_WIN32_FONT_MAP           (pango_win32_font_map_get_type ())
-#define PANGO_WIN32_FONT_MAP(object)        (G_TYPE_CHECK_INSTANCE_CAST ((object), PANGO_TYPE_WIN32_FONT_MAP, PangoWin32FontMap))
-#define PANGO_WIN32_IS_FONT_MAP(object)     (G_TYPE_CHECK_INSTANCE_TYPE ((object), PANGO_TYPE_WIN32_FONT_MAP))
-
-typedef struct _PangoWin32FontMap      PangoWin32FontMap;
 typedef struct _PangoWin32Family       PangoWin32Family;
 typedef struct _PangoWin32SizeInfo     PangoWin32SizeInfo;
 
 /* Number of freed fonts */
 #define MAX_FREED_FONTS 16
 
-struct _PangoWin32FontMap
-{
-  PangoFontMap parent_instance;
-
-  PangoWin32FontCache *font_cache;
-  GQueue *freed_fonts;
-
-  /* Map Pango family names to PangoWin32Family structs */
-  GHashTable *families;
-
-  /* Map LOGFONTS (taking into account only the lfFaceName, lfItalic
-   * and lfWeight fields) to PangoWin32SizeInfo structs.
-   */
-  GHashTable *size_infos;
-
-  int n_fonts;
-
-  double resolution;		/* (points / pixel) * PANGO_SCALE */
-};
-
-struct _PangoWin32FontMapClass
-{
-  PangoFontMapClass parent_class;
-};
-
 struct _PangoWin32Family
 {
   PangoFontFamily parent_instance;
@@ -99,9 +70,6 @@ static void       pango_win32_face_list_sizes        (PangoFontFace  *face,
                                                       int           **sizes,
                                                       int            *n_sizes);
 
-static void       pango_win32_font_map_init          (PangoWin32FontMap       *fontmap);
-static void       pango_win32_font_map_class_init    (PangoFontMapClass       *class);
-
 static void       pango_win32_font_map_finalize      (GObject                      *object);
 static PangoFont *pango_win32_font_map_load_font     (PangoFontMap                 *fontmap,
 						      PangoContext                 *context,
@@ -110,103 +78,78 @@ static void       pango_win32_font_map_list_families (PangoFontMap
 						       PangoFontFamily            ***families,
 						       int                          *n_families);
   
+static PangoFont *pango_win32_font_map_real_find_font (PangoWin32FontMap          *win32fontmap,
+						       PangoContext               *context,
+						       PangoWin32Face             *face,
+						       const PangoFontDescription *description);
+
 static void       pango_win32_fontmap_cache_clear    (PangoWin32FontMap            *win32fontmap);
 
 static void       pango_win32_insert_font            (PangoWin32FontMap            *fontmap,
 						      LOGFONT                      *lfp);
 
-static PangoFontClass *font_map_parent_class;	/* Parent class structure for PangoWin32FontMap */
+static PangoWin32FontMap *default_fontmap = NULL;
 
-static GType
-pango_win32_font_map_get_type (void)
+G_DEFINE_TYPE (PangoWin32FontMap, pango_win32_font_map, PANGO_TYPE_FONT_MAP)
+
+#define TOLOWER(c) \
+  (((c) >= 'A' && (c) <= 'Z') ? (c) - 'A' + 'a' : (c))
+
+static guint
+case_insensitive_hash (const char *key)
 {
-  static GType object_type = 0;
+  const char *p = key;
+  guint h = TOLOWER (*p);
 
-  if (!object_type)
+  if (h)
     {
-      static const GTypeInfo object_info =
-      {
-        sizeof (PangoFontMapClass),
-        (GBaseInitFunc) NULL,
-        (GBaseFinalizeFunc) NULL,
-        (GClassInitFunc) pango_win32_font_map_class_init,
-        NULL,           /* class_finalize */
-        NULL,           /* class_data */
-        sizeof (PangoWin32FontMap),
-        0,              /* n_preallocs */
-        (GInstanceInitFunc) pango_win32_font_map_init,
-      };
-      
-      object_type = g_type_register_static (PANGO_TYPE_FONT_MAP,
-                                            "PangoWin32FontMap",
-                                            &object_info, 0);
+      for (p += 1; *p != '\0'; p++)
+	h = (h << 5) - h + TOLOWER (*p);
     }
-  
-  return object_type;
+
+  return h;
+}
+
+static gboolean
+case_insensitive_equal (const char *key1,
+			const char *key2)
+{
+  return (g_ascii_strcasecmp (key1, key2) == 0);
 }
 
 /* A hash function for LOGFONTs that takes into consideration only
  * those fields that indicate a specific .ttf file is in use:
  * lfFaceName, lfItalic and lfWeight. Dunno how correct this is.
  */
-
 static guint
-logfont_nosize_hash (gconstpointer v)
+logfont_nosize_hash (const LOGFONT *lfp)
 {
-  const LOGFONT *lfp = v;
-  gchar *face = g_ascii_strdown (lfp->lfFaceName, -1);
-  guint result = g_str_hash (face) + lfp->lfItalic + lfp->lfWeight;
-
-  g_free (face);
-
-  return result;
+  return case_insensitive_hash (lfp->lfFaceName) + lfp->lfItalic + lfp->lfWeight;
 }
 
 /* Ditto comparison function */
 static gboolean
-logfont_nosize_equal (gconstpointer v1,
-		      gconstpointer v2)
+logfont_nosize_equal (const LOGFONT *lfp1,
+		      const LOGFONT *lfp2)
 {
-  const LOGFONT *lfp1 = v1, *lfp2 = v2;
-
-  return (g_ascii_strcasecmp (lfp1->lfFaceName, lfp2->lfFaceName) == 0
+  return (case_insensitive_equal (lfp1->lfFaceName, lfp2->lfFaceName)
 	  && lfp1->lfItalic == lfp2->lfItalic
 	  && lfp1->lfWeight == lfp2->lfWeight);
 }
   
-static void 
-pango_win32_font_map_init (PangoWin32FontMap *win32fontmap)
-{
-  win32fontmap->families = g_hash_table_new (g_str_hash, g_str_equal);
-  win32fontmap->size_infos =
-    g_hash_table_new (logfont_nosize_hash, logfont_nosize_equal);
-  win32fontmap->n_fonts = 0;
-}
-
-static void
-pango_win32_font_map_class_init (PangoFontMapClass *class)
-{
-  GObjectClass *object_class = G_OBJECT_CLASS (class);
-  
-  font_map_parent_class = g_type_class_peek_parent (class);
-  
-  object_class->finalize = pango_win32_font_map_finalize;
-  class->load_font = pango_win32_font_map_load_font;
-  class->list_families = pango_win32_font_map_list_families;
-  class->shape_engine_type = PANGO_RENDER_TYPE_WIN32;
-  
-  pango_win32_get_dc ();
-}
-
-static PangoWin32FontMap *fontmap = NULL;
-
 static int CALLBACK
 pango_win32_inner_enum_proc (LOGFONT    *lfp,
 			     TEXTMETRIC *metrics,
 			     DWORD       fontType,
 			     LPARAM      lParam)
 {
-  pango_win32_insert_font (fontmap, lfp);
+  PangoWin32FontMap *win32fontmap = (PangoWin32FontMap *)lParam;
+  
+  /* Windows generates synthetic vertical writing versions of East
+   * Asian fonts with @ prepended to their name, ignore them.
+   */
+  if (lfp->lfFaceName[0] != '@')
+    pango_win32_insert_font (win32fontmap, lfp);
 
   return 1;
 }
@@ -219,12 +162,13 @@ pango_win32_enum_proc (LOGFONT    *lfp,
 {
   LOGFONT lf;
 
+  PING(("%s", lfp->lfFaceName));
+
   if (fontType != TRUETYPE_FONTTYPE)
     return 1;
 
   lf = *lfp;
 
-  PING(("%s", lf.lfFaceName));
   EnumFontFamiliesExA (pango_win32_hdc, &lf,
 		       (FONTENUMPROC) pango_win32_inner_enum_proc,
 		       lParam, 0);
@@ -232,6 +176,47 @@ pango_win32_enum_proc (LOGFONT    *lfp,
   return 1;
 }
 
+static void 
+pango_win32_font_map_init (PangoWin32FontMap *win32fontmap)
+{
+  LOGFONT logfont;
+
+  win32fontmap->families = g_hash_table_new ((GHashFunc) case_insensitive_hash,
+					     (GEqualFunc) case_insensitive_equal);
+  win32fontmap->size_infos =
+    g_hash_table_new ((GHashFunc) logfont_nosize_hash, (GEqualFunc) logfont_nosize_equal);
+  win32fontmap->n_fonts = 0;
+  
+  win32fontmap->font_cache = pango_win32_font_cache_new ();
+  win32fontmap->freed_fonts = g_queue_new ();
+
+  memset (&logfont, 0, sizeof (logfont));
+  logfont.lfCharSet = DEFAULT_CHARSET;
+  EnumFontFamiliesExA (pango_win32_hdc, &logfont, (FONTENUMPROC) pango_win32_enum_proc, 
+		       (LPARAM)win32fontmap, 0);
+
+  win32fontmap->resolution = (PANGO_SCALE / (double) GetDeviceCaps (pango_win32_hdc, LOGPIXELSY)) * 72.0;
+}
+
+static void
+pango_win32_font_map_class_init (PangoWin32FontMapClass *class)
+{
+  GObjectClass *object_class = G_OBJECT_CLASS (class);
+  PangoFontMapClass *fontmap_class = PANGO_FONT_MAP_CLASS (class);
+  int i;
+
+  class->find_font = pango_win32_font_map_real_find_font;
+  object_class->finalize = pango_win32_font_map_finalize;
+  fontmap_class->load_font = pango_win32_font_map_load_font;
+  fontmap_class->list_families = pango_win32_font_map_list_families;
+  fontmap_class->shape_engine_type = PANGO_RENDER_TYPE_WIN32;
+  
+  pango_win32_get_dc ();
+
+  for (i = 0; _pango_included_win32_modules[i].list; i++)
+    pango_module_register (&_pango_included_win32_modules[i]);
+}
+
 /**
  * pango_win32_font_map_for_display:
  *
@@ -239,31 +224,20 @@ pango_win32_enum_proc (LOGFONT    *lfp,
  * not be freed. If the font map is no longer needed, it can
  * be released with pango_win32_shutdown_display().
  *
- * Returns: a #PangoFontMap.
+ * Return value: a #PangoFontMap.
  **/
 PangoFontMap *
 pango_win32_font_map_for_display (void)
 {
-  LOGFONT logfont;
-
   /* Make sure that the type system is initialized */
   g_type_init ();
   
-  if (fontmap != NULL)
-    return PANGO_FONT_MAP (fontmap);
+  if (default_fontmap != NULL)
+    return PANGO_FONT_MAP (default_fontmap);
 
-  fontmap = g_object_new (PANGO_TYPE_WIN32_FONT_MAP, NULL);
+  default_fontmap = g_object_new (PANGO_TYPE_WIN32_FONT_MAP, NULL);
   
-  fontmap->font_cache = pango_win32_font_cache_new ();
-  fontmap->freed_fonts = g_queue_new ();
-
-  memset (&logfont, 0, sizeof (logfont));
-  logfont.lfCharSet = DEFAULT_CHARSET;
-  EnumFontFamiliesExA (pango_win32_hdc, &logfont, (FONTENUMPROC) pango_win32_enum_proc, 0, 0);
-
-  fontmap->resolution = (PANGO_SCALE / (double) GetDeviceCaps (pango_win32_hdc, LOGPIXELSY)) * 72.0;
-
-  return PANGO_FONT_MAP (fontmap);
+  return PANGO_FONT_MAP (default_fontmap);
 }
 
 /**
@@ -274,12 +248,12 @@ pango_win32_font_map_for_display (void)
 void
 pango_win32_shutdown_display (void)
 {
-  if (fontmap)
+  if (default_fontmap)
     {
-      pango_win32_fontmap_cache_clear (fontmap);
-      g_object_unref (fontmap);
+      pango_win32_fontmap_cache_clear (default_fontmap);
+      g_object_unref (default_fontmap);
 
-      fontmap = NULL;
+      default_fontmap = NULL;
     }
 }
 
@@ -293,7 +267,7 @@ pango_win32_font_map_finalize (GObject *object)
   
   pango_win32_font_cache_free (win32fontmap->font_cache);
 
-  G_OBJECT_CLASS (font_map_parent_class)->finalize (object);
+  G_OBJECT_CLASS (pango_win32_font_map_parent_class)->finalize (object);
 }
 
 /*
@@ -367,7 +341,7 @@ pango_win32_family_get_type (void)
       };
       
       object_type = g_type_register_static (PANGO_TYPE_FONT_FAMILY,
-                                            "PangoWin32Family",
+                                            I_("PangoWin32Family"),
                                             &object_info, 0);
     }
   
@@ -444,15 +418,13 @@ pango_win32_font_map_load_font (PangoFontMap               *fontmap,
   PangoWin32Family *win32family;
   PangoFont *result = NULL;
   GSList *tmp_list;
-  gchar *name;
 
   g_return_val_if_fail (description != NULL, NULL);
   
-  name = g_ascii_strdown (pango_font_description_get_family (description), -1);
+  PING(("name=%s", pango_font_description_get_family (description)));
 
-  PING(("name=%s", name));
-
-  win32family = g_hash_table_lookup (win32fontmap->families, name);
+  win32family = g_hash_table_lookup (win32fontmap->families,
+				     pango_font_description_get_family (description));
   if (win32family)
     {
       PangoWin32Face *best_match = NULL;
@@ -472,50 +444,59 @@ pango_win32_font_map_load_font (PangoFontMap               *fontmap,
 	}
 
       if (best_match)
-	{
-	  GSList *tmp_list = best_match->cached_fonts;
-	  gint size = pango_font_description_get_size (description);
-
-	  PING(("got best match:%s size=%d",best_match->logfont.lfFaceName,size));
-
-	  while (tmp_list)
-	    {
-	      PangoWin32Font *win32font = tmp_list->data;
-	      if (win32font->size == size)
-		{
-		  PING (("size matches"));
-		  result = (PangoFont *)win32font;
-
-		  g_object_ref (result);
-		  if (win32font->in_cache)
-		    pango_win32_fontmap_cache_remove (fontmap, win32font);
-		  
-		  break;
-		}
-	      tmp_list = tmp_list->next;
-	    }
-
-	  if (!result)
-	    {
-	      PangoWin32Font *win32font = pango_win32_font_new (fontmap, &best_match->logfont, size);
-
-	      win32font->fontmap = fontmap;
-	      win32font->win32face = best_match;
-	      best_match->cached_fonts = g_slist_prepend (best_match->cached_fonts, win32font);
-
-	      result = (PangoFont *)win32font;
-	    }
-	}
+	result = PANGO_WIN32_FONT_MAP_GET_CLASS (win32fontmap)->find_font (win32fontmap, context,
+									   best_match,
+									   description);
       else
 	PING(("no best match!"));
     }
 
-  g_free (name);
   return result;
 }
 
+static PangoFont *
+pango_win32_font_map_real_find_font (PangoWin32FontMap          *win32fontmap,
+				     PangoContext               *context,
+				     PangoWin32Face             *face,
+				     const PangoFontDescription *description)
+{
+  PangoFontMap *fontmap = PANGO_FONT_MAP (win32fontmap);
+  PangoWin32Font *win32font;
+  GSList *tmp_list = face->cached_fonts;
+  int size = pango_font_description_get_size (description);
+  
+  if (pango_font_description_get_size_is_absolute (description))
+    size = (int) 0.5 + (size * win32fontmap->resolution) / PANGO_SCALE;
+  
+  PING(("got best match:%s size=%d",face->logfont.lfFaceName,size));
+  
+  while (tmp_list)
+    {
+      win32font = tmp_list->data;
+      if (win32font->size == size)
+	{
+	  PING (("size matches"));
+	  
+	  g_object_ref (win32font);
+	  if (win32font->in_cache)
+	    pango_win32_fontmap_cache_remove (fontmap, win32font);
+	  
+	  return (PangoFont *)win32font;
+	}
+      tmp_list = tmp_list->next;
+    }
+  
+  win32font = pango_win32_font_new (fontmap, &face->logfont, size);
+      
+  win32font->fontmap = fontmap;
+  win32font->win32face = face;
+  face->cached_fonts = g_slist_prepend (face->cached_fonts, win32font);
+  
+  return (PangoFont *)win32font;
+}
+
 static gchar *
-get_family_name (LOGFONT *lfp)
+get_family_name (const LOGFONT *lfp)
 {
   HFONT hfont;
   HFONT oldhfont;
@@ -638,14 +619,74 @@ get_family_name (LOGFONT *lfp)
   return g_locale_to_utf8 (lfp->lfFaceName, -1, NULL, NULL, NULL);
 }
 
-static gchar *
-get_family_name_lowercase (LOGFONT *lfp)
+/**
+ * pango_win32_font_description_from_logfont:
+ * @lfp: a LOGFONT
+ *
+ * Creates a #PangoFontDescription that matches the specified LOGFONT.
+ *
+ * The face name, italicness and weight fields in the LOGFONT are used
+ * to set up the resulting #PangoFontDescription. If the face name in
+ * the LOGFONT contains non-ASCII characters the font is temporarily
+ * loaded (using CreateFontIndirect) and an ASCII (usually English)
+ * name for it is looked up from the font name tables in the font
+ * data. If that doesn't work, the face name is converted from the
+ * system codepage to UTF-8 and that is used.
+ *
+ * Return value: the newly allocated #PangoFontDescription, which
+ *  should be freed using pango_font_desciption_free()
+ *
+ * Since: 1.12
+ */
+PangoFontDescription *
+pango_win32_font_description_from_logfont (const LOGFONT *lfp)
 {
-  gchar *family = get_family_name (lfp);
-  gchar *lower = g_ascii_strdown (family, -1);
+  PangoFontDescription *description;
+  gchar *family;
+  PangoStyle style;
+  PangoVariant variant;
+  PangoWeight weight;
+  PangoStretch stretch;
 
-  g_free (family);
-  return lower;
+  family = get_family_name (lfp);
+
+  if (!lfp->lfItalic)
+    style = PANGO_STYLE_NORMAL;
+  else
+    style = PANGO_STYLE_ITALIC;
+
+  variant = PANGO_VARIANT_NORMAL;
+  
+  /* The PangoWeight values PANGO_WEIGHT_* map exactly do Windows FW_*
+   * values.  Is this on purpose? Quantize the weight to exact
+   * PANGO_WEIGHT_* values. Is this a good idea?
+   */
+  if (lfp->lfWeight == FW_DONTCARE)
+    weight = PANGO_WEIGHT_NORMAL;
+  else if (lfp->lfWeight <= (FW_ULTRALIGHT + FW_LIGHT) / 2)
+    weight = PANGO_WEIGHT_ULTRALIGHT;
+  else if (lfp->lfWeight <= (FW_LIGHT + FW_NORMAL) / 2)
+    weight = PANGO_WEIGHT_LIGHT;
+  else if (lfp->lfWeight <= (FW_NORMAL + FW_BOLD) / 2)
+    weight = PANGO_WEIGHT_NORMAL;
+  else if (lfp->lfWeight <= (FW_BOLD + FW_ULTRABOLD) / 2)
+    weight = PANGO_WEIGHT_BOLD;
+  else if (lfp->lfWeight <= (FW_ULTRABOLD + FW_HEAVY) / 2)
+    weight = PANGO_WEIGHT_ULTRABOLD;
+  else
+    weight = PANGO_WEIGHT_HEAVY;
+
+  /* XXX No idea how to figure out the stretch */
+  stretch = PANGO_STRETCH_NORMAL;
+  
+  description = pango_font_description_new ();
+  pango_font_description_set_family (description, family);
+  pango_font_description_set_style (description, style);
+  pango_font_description_set_weight (description, weight);
+  pango_font_description_set_stretch (description, stretch);
+  pango_font_description_set_variant (description, variant);
+
+  return description;
 }
 
 /* This inserts the given font into the size_infos table. If a SizeInfo
@@ -658,20 +699,22 @@ pango_win32_insert_font (PangoWin32FontMap *win32fontmap,
 			 LOGFONT           *lfp)
 {
   LOGFONT *lfp2 = NULL;
-  PangoFontDescription *description;
-  gchar *family_name;
   PangoWin32Family *font_family;
   PangoWin32Face *win32face;
   PangoWin32SizeInfo *size_info;
-  PangoStyle style;
-  PangoVariant variant;
-  PangoWeight weight;
-  PangoStretch stretch;
   GSList *tmp_list;
   gint i;
 
   PING(("face=%s,charset=%d,it=%d,wt=%ld,ht=%ld",lfp->lfFaceName,lfp->lfCharSet,lfp->lfItalic,lfp->lfWeight,lfp->lfHeight));
   
+  /* Ignore Symbol fonts (which don't have any Unicode mapping
+   * table). We could also be fancy and use the PostScript glyph name
+   * table for such if present, and build a Unicode map by mapping
+   * each PostScript glyph name to Unicode character. Oh well.
+   */
+  if (lfp->lfCharSet == SYMBOL_CHARSET)
+    return;
+
   /* First insert the LOGFONT into the list of LOGFONTs for the typeface name
    */
   size_info = g_hash_table_lookup (win32fontmap->size_infos, lfp);
@@ -702,7 +745,10 @@ pango_win32_insert_font (PangoWin32FontMap *win32fontmap,
 	      rover->lfOrientation == lfp->lfOrientation &&
 	      rover->lfUnderline == lfp->lfUnderline &&
 	      rover->lfStrikeOut == lfp->lfStrikeOut)
-	    return;
+	    {
+	      PING(("already have it"));
+	      return;
+	    }
 	  
 	  tmp_list = tmp_list->next;
 	}
@@ -718,75 +764,23 @@ pango_win32_insert_font (PangoWin32FontMap *win32fontmap,
   
   PING(("g_slist_length(size_info->logfonts)=%d", g_slist_length(size_info->logfonts)));
 
-  family_name = get_family_name_lowercase (lfp);
-
-  /* Convert the LOGFONT into a PangoFontDescription
-   */
-  if (!lfp2->lfItalic)
-    style = PANGO_STYLE_NORMAL;
-  else
-    style = PANGO_STYLE_ITALIC;
-
-  variant = PANGO_VARIANT_NORMAL;
-  
-  /* The PangoWeight values PANGO_WEIGHT_* map exactly do Windows FW_*
-   * values.  Is this on purpose? Quantize the weight to exact
-   * PANGO_WEIGHT_* values. Is this a good idea?
-   */
-  if (lfp2->lfWeight == FW_DONTCARE)
-    weight = PANGO_WEIGHT_NORMAL;
-  else if (lfp2->lfWeight <= (FW_ULTRALIGHT + FW_LIGHT) / 2)
-    weight = PANGO_WEIGHT_ULTRALIGHT;
-  else if (lfp2->lfWeight <= (FW_LIGHT + FW_NORMAL) / 2)
-    weight = PANGO_WEIGHT_LIGHT;
-  else if (lfp2->lfWeight <= (FW_NORMAL + FW_BOLD) / 2)
-    weight = PANGO_WEIGHT_NORMAL;
-  else if (lfp2->lfWeight <= (FW_BOLD + FW_ULTRABOLD) / 2)
-    weight = PANGO_WEIGHT_BOLD;
-  else if (lfp2->lfWeight <= (FW_ULTRABOLD + FW_HEAVY) / 2)
-    weight = PANGO_WEIGHT_ULTRABOLD;
-  else
-    weight = PANGO_WEIGHT_HEAVY;
-
-  /* XXX No idea how to figure out the stretch */
-  stretch = PANGO_STRETCH_NORMAL;
-
-  font_family = pango_win32_get_font_family (win32fontmap, family_name);
-  g_free (family_name);
-
-  tmp_list = font_family->font_entries;
-  while (tmp_list)
-    {
-      win32face = tmp_list->data;
-
-      if (pango_font_description_get_style (win32face->description) == style &&
-	  pango_font_description_get_weight (win32face->description) == weight &&
-	  pango_font_description_get_stretch (win32face->description) == stretch &&
-	  pango_font_description_get_variant (win32face->description) == variant)
-	{
-	  PING(("returning early"));
-	  return;
-	}
-
-      tmp_list = tmp_list->next;
-    }
-
-  description = pango_font_description_new ();
-  pango_font_description_set_family_static (description, font_family->family_name);
-  pango_font_description_set_style (description, style);
-  pango_font_description_set_weight (description, weight);
-  pango_font_description_set_stretch (description, stretch);
-  pango_font_description_set_variant (description, variant);
-
   win32face = g_object_new (PANGO_WIN32_TYPE_FACE, NULL);
-  win32face->description = description;
+  win32face->description = pango_win32_font_description_from_logfont (lfp2);
+
   win32face->cached_fonts = NULL;
   
   for (i = 0; i < PANGO_WIN32_N_COVERAGES; i++)
      win32face->coverages[i] = NULL;
   win32face->logfont = *lfp;
-  win32face->unicode_table = NULL;
+  win32face->cmap_format = 0;
+  win32face->cmap = NULL;
+
+  font_family =
+    pango_win32_get_font_family (win32fontmap,
+				 pango_font_description_get_family (win32face->description));
   font_family->font_entries = g_slist_append (font_family->font_entries, win32face);
+  PING(("g_slist_length(font_family->font_entries)=%d", g_slist_length(font_family->font_entries)));
+
   win32fontmap->n_fonts++;
 
 #if 1 /* Thought pango.aliases would make this code unnecessary, but no. */
@@ -987,7 +981,7 @@ pango_win32_face_get_type (void)
       };
       
       object_type = g_type_register_static (PANGO_TYPE_FONT_FACE,
-                                            "PangoWin32Face",
+                                            I_("PangoWin32Face"),
                                             &object_info, 0);
     }
   
@@ -1021,7 +1015,7 @@ pango_win32_font_entry_remove (PangoWin32Face *face,
  *
  * Obtains the font cache associated with the given font map.
  *
- * Returns: the #PangoWin32FontCache of @font_map.
+ * Return value: the #PangoWin32FontCache of @font_map.
  **/
 PangoWin32FontCache *
 pango_win32_font_map_get_font_cache (PangoFontMap *font_map)
@@ -1054,17 +1048,10 @@ pango_win32_fontmap_cache_remove (PangoFontMap   *fontmap,
 				  PangoWin32Font *win32font)
 {
   PangoWin32FontMap *win32fontmap = PANGO_WIN32_FONT_MAP (fontmap);
+  GList *link = g_queue_find (win32fontmap->freed_fonts, win32font);
 
-  GList *link = g_list_find (win32fontmap->freed_fonts->head, win32font);
-  if (link == win32fontmap->freed_fonts->tail)
-    {
-      win32fontmap->freed_fonts->tail = win32fontmap->freed_fonts->tail->prev;
-      if (win32fontmap->freed_fonts->tail)
-	win32fontmap->freed_fonts->tail->next = NULL;
-    }
-  
-  win32fontmap->freed_fonts->head = g_list_delete_link (win32fontmap->freed_fonts->head, link);
-  win32fontmap->freed_fonts->length--;
+  if (link)
+    g_queue_delete_link (win32fontmap->freed_fonts, link);
   win32font->in_cache = FALSE;
 
   g_object_unref (win32font);
@@ -1074,8 +1061,6 @@ static void
 pango_win32_fontmap_cache_clear (PangoWin32FontMap *win32fontmap)
 {
   g_list_foreach (win32fontmap->freed_fonts->head, (GFunc)g_object_unref, NULL);
-  g_list_free (win32fontmap->freed_fonts->head);
-  win32fontmap->freed_fonts->head = NULL;
-  win32fontmap->freed_fonts->tail = NULL;
-  win32fontmap->freed_fonts->length = 0;
+  g_queue_free (win32fontmap->freed_fonts);
+  win32fontmap->freed_fonts = g_queue_new ();
 }
