@@ -39,10 +39,8 @@ static void fail (const char *format, ...)
   str = g_strdup_vprintf (format, args);
   va_end (args);
 
-  fprintf (stderr, "Error: %s\n", str);
+  fprintf (stderr, "Error: %s\n", format);
   
-  g_free (str);
-
   exit (1);
 }
 
